/*
 * Name:    d_network_ainput.c
 *
 * Purpose: MX network analog input device driver.
 *
 * Author:  William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 1999, 2001, 2003-2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "mx_util.h"
#include "mx_driver.h"
#include "mx_net.h"
#include "mx_analog_input.h"
#include "d_network_ainput.h"

MX_RECORD_FUNCTION_LIST mxd_network_ainput_record_function_list = {
	NULL,
	mxd_network_ainput_create_record_structures,
	mx_analog_input_finish_record_initialization
};

MX_ANALOG_INPUT_FUNCTION_LIST mxd_network_ainput_analog_input_function_list = {
	mxd_network_ainput_read,
	mxd_network_ainput_get_dark_current,
	mxd_network_ainput_set_dark_current
};

MX_RECORD_FIELD_DEFAULTS mxd_network_ainput_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MX_DOUBLE_ANALOG_INPUT_STANDARD_FIELDS,
	MX_ANALOG_INPUT_STANDARD_FIELDS,
	MXD_NETWORK_AINPUT_STANDARD_FIELDS
};

long mxd_network_ainput_num_record_fields
		= sizeof( mxd_network_ainput_record_field_defaults )
			/ sizeof( mxd_network_ainput_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxd_network_ainput_rfield_def_ptr
			= &mxd_network_ainput_record_field_defaults[0];

/* ===== */

MX_EXPORT mx_status_type
mxd_network_ainput_create_record_structures( MX_RECORD *record )
{
        static const char fname[] =
			"mxd_network_ainput_create_record_structures()";

        MX_ANALOG_INPUT *analog_input;
        MX_NETWORK_AINPUT *network_ainput;

        /* Allocate memory for the necessary structures. */

        analog_input = (MX_ANALOG_INPUT *) malloc(sizeof(MX_ANALOG_INPUT));

        if ( analog_input == (MX_ANALOG_INPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_ANALOG_INPUT structure." );
        }

        network_ainput = (MX_NETWORK_AINPUT *)
				malloc( sizeof(MX_NETWORK_AINPUT) );

        if ( network_ainput == (MX_NETWORK_AINPUT *) NULL ) {
                return mx_error( MXE_OUT_OF_MEMORY, fname,
                "Can't allocate memory for MX_NETWORK_AINPUT structure." );
        }

        /* Now set up the necessary pointers. */

        record->record_class_struct = analog_input;
        record->record_type_struct = network_ainput;
        record->class_specific_function_list
			= &mxd_network_ainput_analog_input_function_list;

        analog_input->record = record;

	/* Raw analog input values are stored as doubles. */

	analog_input->subclass = MXT_AIN_DOUBLE;

        return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_read( MX_ANALOG_INPUT *ainput )
{
	static const char fname[] = "mxd_network_ainput_read()";

	MX_NETWORK_AINPUT *network_ainput;
	char rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	double raw_value;
	mx_status_type mx_status;

	network_ainput = (MX_NETWORK_AINPUT *)
				ainput->record->record_type_struct;

	if ( network_ainput == NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_NETWORK_AINPUT pointer for analog input '%s' is NULL",
			ainput->record->name );
	}

	sprintf( rfname, "%s.value", network_ainput->remote_record_name );

	mx_status = mx_get_by_name( network_ainput->server_record,
				rfname, MXFT_DOUBLE, &raw_value );

	ainput->raw_value.double_value = raw_value;

	return mx_status;
}

MX_EXPORT mx_status_type
mxd_network_ainput_get_dark_current( MX_ANALOG_INPUT *ainput )
{
	static const char fname[] = "mxd_network_ainput_get_dark_current()";

	MX_NETWORK_AINPUT *network_ainput;
	char rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	double dark_current;
	mx_status_type mx_status;

	/* If the database is configured such that the server maintains
	 * the dark current rather than the client, then ask the server
	 * for the value.  Otherwise, just return our locally cached value.
	 */

	if (ainput->analog_input_flags & MXF_AIN_SERVER_SUBTRACTS_DARK_CURRENT)
	{
		network_ainput = (MX_NETWORK_AINPUT *)
				ainput->record->record_type_struct;

		if ( network_ainput == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		    "MX_NETWORK_AINPUT pointer for analog input '%s' is NULL",
				ainput->record->name );
		}

		sprintf( rfname, "%s.dark_current",
				network_ainput->remote_record_name );

		mx_status = mx_get_by_name(
				network_ainput->server_record,
				rfname, MXFT_DOUBLE, &dark_current );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;

		ainput->dark_current = dark_current;
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxd_network_ainput_set_dark_current( MX_ANALOG_INPUT *ainput )
{
	static const char fname[] = "mxd_network_ainput_set_dark_current()";

	MX_NETWORK_AINPUT *network_ainput;
	char rfname[MXU_RECORD_FIELD_NAME_LENGTH + 1];
	double dark_current;
	mx_status_type mx_status;

	/* If the database is configured such that the server maintains
	 * the dark current rather than the client, then send the value
	 * to the server.  Otherwise, just cache it locally.
	 */

	if (ainput->analog_input_flags & MXF_AIN_SERVER_SUBTRACTS_DARK_CURRENT)
	{
		network_ainput = (MX_NETWORK_AINPUT *)
				ainput->record->record_type_struct;

		if ( network_ainput == NULL ) {
			return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		    "MX_NETWORK_AINPUT pointer for analog input '%s' is NULL",
				ainput->record->name );
		}

		dark_current = ainput->dark_current;

		sprintf( rfname, "%s.dark_current",
				network_ainput->remote_record_name );

		mx_status = mx_put_by_name(
				network_ainput->server_record,
				rfname, MXFT_DOUBLE, &dark_current );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}

	return MX_SUCCESSFUL_RESULT;
}

