/*
 * Name:    mx_net.c
 *
 * Purpose: Client side part of Spec network protocol.
 *
 * Author:  William Lavender
 *
 *-------------------------------------------------------------------------
 *
 * Copyright 2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_socket.h"
#include "mx_spec.h"
#include "n_spec.h"

#define SPEC_DEBUG		TRUE

#define SPEC_DEBUG_TIMING	FALSE

#if SPEC_DEBUG_TIMING
#include "mx_hrt_debug.h"
#endif

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_spec_send_message( MX_RECORD *spec_server_record,
                        int spec_command_code,
                        int spec_datatype,
                        unsigned spec_array_data_rows,
                        unsigned spec_array_data_cols,
                        unsigned spec_data_length,
                        char *spec_property_name,
                        void *data_pointer )
{
	static const char fname[] = "mx_spec_send_message()";

	MX_SPEC_SERVER *spec_server;
	struct svr_head message_header;
	size_t message_header_size;
	mx_status_type mx_status;

	if ( spec_server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	spec_server = (MX_SPEC_SERVER *) spec_server_record->record_type_struct;

	if ( spec_server == (MX_SPEC_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SPEC_SERVER pointer for record '%s' is NULL.",
			spec_server_record->name );
	}

	message_header_size = sizeof(struct svr_head);

	message_header.magic = SV_SPEC_MAGIC;
	message_header.vers = SV_VERSION;
	message_header.size = message_header_size;
	message_header.sn = 0;
	message_header.sec = 0;
	message_header.usec = 0;
	message_header.cmd = spec_command_code;
	message_header.type = spec_datatype;
	message_header.rows = spec_array_data_rows;
	message_header.cols = spec_array_data_cols;
	message_header.len = spec_data_length;

	mx_strncpy( message_header.name, spec_property_name, SV_NAME_LEN );

	/* Send the message header. */

	mx_status = mx_socket_send( spec_server->socket,
					&message_header,
					message_header_size );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Now send the message data. */

	mx_status = mx_socket_send( spec_server->socket,
					data_pointer,
					spec_data_length );

	return mx_status;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_spec_receive_message( MX_RECORD *spec_server_record,
                        int *spec_command_code,
                        int *spec_datatype,
                        unsigned *spec_array_data_rows,
                        unsigned *spec_array_data_cols,
                        unsigned maximum_data_length,
                        unsigned *spec_data_length,
                        char *spec_property_name,
                        void *data_pointer )
{
	static const char fname[] = "mx_spec_receive_message()";

	MX_SPEC_SERVER *spec_server;
	struct svr_head message_header;
	size_t expected_message_header_size;
	mx_status_type mx_status;

	if ( spec_server_record == (MX_RECORD *) NULL ) {
		return mx_error( MXE_NULL_ARGUMENT, fname,
		"The MX_RECORD pointer passed was NULL." );
	}

	spec_server = (MX_SPEC_SERVER *) spec_server_record->record_type_struct;

	if ( spec_server == (MX_SPEC_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_SPEC_SERVER pointer for record '%s' is NULL.",
			spec_server_record->name );
	}

	expected_message_header_size = sizeof(struct svr_head);

	/* Receive the header of the message. */

	mx_status = mx_socket_receive( spec_server->socket,
					&message_header,
					expected_message_header_size,
					NULL, NULL, 0 );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	if ( message_header.magic != SV_SPEC_MAGIC ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"Wrong Spec magic number %#x found in header from Spec server '%s'.",
			message_header.magic, spec_server_record->name );
	}
	if ( message_header.vers != SV_VERSION ) {
		mx_warning( "Spec server '%s' is using protocol version %d, "
				"but we were expecting version %d.",
				spec_server_record->name,
				message_header.vers,
				SV_VERSION );
	}
	if ( message_header.size != expected_message_header_size ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
	"The message header size %u for the header sent by Spec server '%s' "
	"does not match the expected value of %u.",
			message_header.size, spec_server_record->name,
			(unsigned int) expected_message_header_size );
	}

	if ( spec_command_code != NULL ) {
		*spec_command_code = message_header.cmd;
	}
	if ( spec_datatype != NULL ) {
		*spec_datatype = message_header.type;
	}
	if ( spec_array_data_rows != NULL ) {
		*spec_array_data_rows = message_header.rows;
	}
	if ( spec_array_data_cols != NULL ) {
		*spec_array_data_cols = message_header.cols;
	}
	if ( spec_data_length != NULL ) {
		*spec_data_length = message_header.len;
	}
	if ( spec_property_name != NULL ) {
		mx_strncpy( spec_property_name,
			message_header.name, SV_NAME_LEN );
	}

	if ( message_header.len > maximum_data_length ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The message data size of %u bytes from Spec server '%s' "
		"is larger than the maximum length %u specified for "
		"this call.", message_header.len,
			spec_server_record->name,
			maximum_data_length );
	}

	/* Now receive the data for the message. */

	mx_status = mx_socket_receive( spec_server->socket,
					data_pointer,
					message_header.len,
					NULL, NULL, 0 );

	return mx_status;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_spec_command( MX_RECORD *spec_server_record,
			char *command,
			char *response,
			size_t response_buffer_length )
{
	int command_code, received_command_code;
	int received_datatype;
	unsigned received_data_rows, received_data_cols;
	unsigned received_data_length;
	char received_property_name[SV_NAME_LEN];
	mx_status_type mx_status;

	if ( response != NULL ) {
		command_code = SV_CMD_WITH_RETURN;
	} else {
		command_code = SV_CMD;
	}

	/* Send the command to Spec. */

	mx_status = mx_spec_send_message( spec_server_record,
					command_code,
					SV_STRING,
					0, 0,
					strlen( command ) + 1, "",
					command );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;
					
	/* If the response pointer is NULL, then no response is expected. */

	if ( response == NULL ) {
		return MX_SUCCESSFUL_RESULT;
	}

	/* If we get here, a response from Spec is expected. */

	mx_status = mx_spec_receive_message( spec_server_record,
					&received_command_code,
					&received_datatype,
					&received_data_rows,
					&received_data_cols,
					response_buffer_length,
					&received_data_length,
					received_property_name,
					response );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_spec_get_array( MX_RECORD *spec_server_record,
			char *property_name,
			long spec_datatype,
			long num_dimensions,
			long *dimension_array,
			void *value )
{
	static const char fname[] = "mx_spec_get_array()";

	unsigned data_rows, data_columns, data_length;
	unsigned received_data_rows, received_data_columns;
	unsigned received_data_length;
	int received_command_code, received_datatype;
	char received_property_name[SV_NAME_LEN];
	mx_status_type mx_status;

	if ( ( num_dimensions < 1 ) || ( num_dimensions > 2 ) ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The requested number of dimensions %ld for Spec server '%s' "
		"is outside the allowed range of values 1 or 2.",
			num_dimensions, spec_server_record->name );
	}

	switch( spec_datatype ) {
	case SV_DOUBLE:
		return mx_error( MXE_UNSUPPORTED, fname,
"Spec server '%s' currently cannot send data using the SV_DOUBLE datatype.",
			spec_server_record->name );
		break;
	case SV_STRING:
		if ( num_dimensions != 1 ) {
			return mx_error( MXE_UNSUPPORTED, fname,
			"Only 1-dimensional strings are supported by Spec "
			"server '%s' for datatype SV_STRING.",
				spec_server_record->name );
		}
		
		data_rows = 1;
		data_columns = dimension_array[0];
		data_length = data_columns;
		break;
	case SV_ERROR:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
    "A Spec client cannot transmit an SV_ERROR message to Spec server '%s'.",
			spec_server_record->name );
		break;
	case SV_ASSOC:
	case SV_ARR_DOUBLE:
	case SV_ARR_FLOAT:
	case SV_ARR_LONG:
	case SV_ARR_ULONG:
	case SV_ARR_SHORT:
	case SV_ARR_USHORT:
	case SV_ARR_CHAR:
	case SV_ARR_UCHAR:
	case SV_ARR_STRING:
		if ( num_dimensions == 1 ) {
			data_rows = 1;
			data_columns = dimension_array[0];
			data_length = data_columns;
		} else {
			data_rows = dimension_array[0];
			data_columns = dimension_array[1];
			data_length = data_rows * data_columns;
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Unrecognized Spec datatype %ld for Spec server '%s', property '%s'.",
			spec_datatype, spec_server_record->name,
			property_name );
		break;
	}

	/* Multiply the data length by the size in bytes of an individual
	 * data element.
	 */

	switch( spec_datatype ) {
	case SV_STRING:
	case SV_ASSOC:
	case SV_ARR_STRING:
	case SV_ARR_CHAR:
	case SV_ARR_UCHAR:
		data_length *= 1;
		break;
	case SV_ARR_SHORT:
	case SV_ARR_USHORT:
		data_length *= 2;
		break;
	case SV_ARR_LONG:
	case SV_ARR_ULONG:
	case SV_ARR_FLOAT:
		data_length *= 4;
		break;
	case SV_ARR_DOUBLE:
		data_length *= 8;
		break;
	}

	/* Send the command to Spec. */

	mx_status = mx_spec_send_message( spec_server_record,
					SV_CHAN_READ, spec_datatype,
					data_rows, data_columns, data_length,
					property_name, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Wait for the response. */

	mx_status = mx_spec_receive_message( spec_server_record,
					&received_command_code,
					&received_datatype,
					&received_data_rows,
					&received_data_columns,
					data_length,
					&received_data_length,
					received_property_name,
					value );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	return MX_SUCCESSFUL_RESULT;
}

/* ====================================================================== */

MX_EXPORT mx_status_type
mx_spec_put_array( MX_RECORD *spec_server_record,
			char *property_name,
			long spec_datatype,
			long num_dimensions,
			long *dimension_array,
			void *value )
{
	static const char fname[] = "mx_spec_put_array()";

	unsigned data_rows, data_columns, data_length;
	mx_status_type mx_status;

	if ( ( num_dimensions < 1 ) || ( num_dimensions > 2 ) ) {
		return mx_error( MXE_WOULD_EXCEED_LIMIT, fname,
		"The requested number of dimensions %ld for Spec server '%s' "
		"is outside the allowed range of values 1 or 2.",
			num_dimensions, spec_server_record->name );
	}

	switch( spec_datatype ) {
	case SV_DOUBLE:
		return mx_error( MXE_UNSUPPORTED, fname,
"Spec server '%s' currently cannot receive data using the SV_DOUBLE datatype.",
			spec_server_record->name );
		break;
	case SV_STRING:
		if ( num_dimensions != 1 ) {
			return mx_error( MXE_UNSUPPORTED, fname,
			"Only 1-dimensional strings are supported by Spec "
			"server '%s' for datatype SV_STRING.",
				spec_server_record->name );
		}
		
		data_rows = 1;
		data_columns = dimension_array[0];
		data_length = data_columns;
		break;
	case SV_ERROR:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
    "A Spec client cannot transmit an SV_ERROR message to Spec server '%s'.",
			spec_server_record->name );
		break;
	case SV_ASSOC:
	case SV_ARR_DOUBLE:
	case SV_ARR_FLOAT:
	case SV_ARR_LONG:
	case SV_ARR_ULONG:
	case SV_ARR_SHORT:
	case SV_ARR_USHORT:
	case SV_ARR_CHAR:
	case SV_ARR_UCHAR:
	case SV_ARR_STRING:
		if ( num_dimensions == 1 ) {
			data_rows = 1;
			data_columns = dimension_array[0];
			data_length = data_columns;
		} else {
			data_rows = dimension_array[0];
			data_columns = dimension_array[1];
			data_length = data_rows * data_columns;
		}
		break;
	default:
		return mx_error( MXE_ILLEGAL_ARGUMENT, fname,
	"Unrecognized Spec datatype %ld for Spec server '%s', property '%s'.",
			spec_datatype, spec_server_record->name,
			property_name );
		break;
	}

	/* Multiply the data length by the size in bytes of an individual
	 * data element.
	 */

	switch( spec_datatype ) {
	case SV_STRING:
	case SV_ASSOC:
	case SV_ARR_STRING:
	case SV_ARR_CHAR:
	case SV_ARR_UCHAR:
		data_length *= 1;
		break;
	case SV_ARR_SHORT:
	case SV_ARR_USHORT:
		data_length *= 2;
		break;
	case SV_ARR_LONG:
	case SV_ARR_ULONG:
	case SV_ARR_FLOAT:
		data_length *= 4;
		break;
	case SV_ARR_DOUBLE:
		data_length *= 8;
		break;
	}

	/* Send the command to Spec. */

	mx_status = mx_spec_send_message( spec_server_record,
					SV_CHAN_SEND, spec_datatype,
					data_rows, data_columns, data_length,
					property_name, value );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_spec_get_string( MX_RECORD *spec_server_record,
			char *property_name,
			char *string_value,
			size_t max_string_length )
{
	int returned_command_code, returned_datatype;
	unsigned returned_data_rows, returned_data_cols, returned_data_length;
	char returned_property_name[SV_NAME_LEN];
	mx_status_type mx_status;

	/* Request the string. */

	mx_status = mx_spec_send_message( spec_server_record,
						SV_CHAN_READ,
						SV_STRING,
						0, 0, 0,
						property_name, NULL );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	/* Receive the string. */

	mx_status = mx_spec_receive_message( spec_server_record,
						&returned_command_code,
						&returned_datatype,
						&returned_data_rows,
						&returned_data_cols,
						max_string_length,
						&returned_data_length,
						returned_property_name,
						string_value );

	return mx_status;
}

MX_EXPORT mx_status_type
mx_spec_put_string( MX_RECORD *spec_server_record,
			char *property_name,
			char *string_value )
{
	mx_status_type mx_status;

	/* Send the string. */

	mx_status = mx_spec_send_message( spec_server_record,
						SV_CHAN_SEND,
						SV_STRING,
						0, 0,
						strlen(string_value) + 1,
						property_name,
						string_value );

	return mx_status;
}

/*
 * WARNING: The following functions use MX datatypes rather than Spec datatypes.
 */

MX_EXPORT mx_status_type
mx_spec_get_number( MX_RECORD *spec_server_record,
			char *property_name,
			long mx_datatype,
			void *value )
{
	static const char fname[] = "mx_spec_get_number()";

	char string_buffer[80];
	char format[20];
	int num_items;
	mx_status_type mx_status;

	mx_status = mx_spec_get_string( spec_server_record,
					property_name,
					string_buffer,
					sizeof(string_buffer) );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	switch( mx_datatype ) {
	case MXFT_CHAR:
	case MXFT_UCHAR:
		strcpy( format, "%c" );
		break;
	case MXFT_SHORT:
		strcpy( format, "%hd" );
		break;
	case MXFT_USHORT:
		strcpy( format, "%hu" );
		break;
	case MXFT_INT:
		strcpy( format, "%d" );
		break;
	case MXFT_UINT:
		strcpy( format, "%u" );
		break;
	case MXFT_LONG:
		strcpy( format, "%ld" );
		break;
	case MXFT_ULONG:
		strcpy( format, "%lu" );
		break;
	case MXFT_HEX:
		strcpy( format, "%lx" );
		break;
	case MXFT_FLOAT:
		strcpy( format, "%g" );
		break;
	case MXFT_DOUBLE:
		strcpy( format, "%lg" );
		break;
	default:
		return mx_error( MXE_UNSUPPORTED, fname,
"Unsupported MX datatype %ld passed for property '%s' of Spec server '%s'.",
			mx_datatype, property_name, spec_server_record->name );
		break;
	}

	num_items = sscanf( string_buffer, format, value );

	if ( num_items != 1 ) {
		return mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Did not find data of type %s in the response from "
		"Spec server '%s' for property '%s'.  Response = '%s'",
			mx_get_field_type_string( mx_datatype ),
			spec_server_record->name, property_name,
			string_buffer );
	}

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mx_spec_put_number( MX_RECORD *spec_server_record,
			char *property_name,
			long mx_datatype,
			void *value )
{
	static const char fname[] = "mx_spec_put_number()";

	char string_buffer[80];
	mx_status_type mx_status;

	switch( mx_datatype ) {
	case MXFT_CHAR:
	case MXFT_UCHAR:
		sprintf( string_buffer, "%c", *((char *) value));
		break;
	case MXFT_SHORT:
		sprintf( string_buffer, "%hd", *((short *) value));
		break;
	case MXFT_USHORT:
		sprintf( string_buffer, "%hu", *((unsigned short *) value));
		break;
	case MXFT_INT:
		sprintf( string_buffer, "%d", *((int *) value));
		break;
	case MXFT_UINT:
		sprintf( string_buffer, "%u", *((unsigned int *) value));
		break;
	case MXFT_LONG:
		sprintf( string_buffer, "%ld", *((long *) value));
		break;
	case MXFT_ULONG:
		sprintf( string_buffer, "%lu", *((unsigned long *) value));
		break;
	case MXFT_HEX:
		sprintf( string_buffer, "%lx", *((unsigned long *) value));
		break;
	case MXFT_FLOAT:
		sprintf( string_buffer, "%g", *((float *) value));
		break;
	case MXFT_DOUBLE:
		sprintf( string_buffer, "%g", *((double *) value));
		break;
	default:
		return mx_error( MXE_UNSUPPORTED, fname,
"Unsupported MX datatype %ld passed for property '%s' of Spec server '%s'.",
			mx_datatype, property_name, spec_server_record->name );
		break;
	}

	mx_status = mx_spec_put_string( spec_server_record,
					property_name,
					string_buffer );

	return mx_status;
}

