/*
 * Name:     n_spec.c
 *
 * Purpose:  Client interface to a Spec server.  Spec is a product of
 *           Certified Scientific Software. (www.certif.com)
 *
 * Author:   William Lavender
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2004 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "mx_util.h"
#include "mx_record.h"
#include "mx_socket.h"
#include "mx_spec.h"
#include "n_spec.h"

MX_RECORD_FUNCTION_LIST mxn_spec_server_record_function_list = {
	NULL,
	mxn_spec_server_create_record_structures,
	mxn_spec_server_finish_record_initialization,
	NULL,
	NULL,
	NULL,
	NULL,
	mxn_spec_server_open,
	mxn_spec_server_close,
	NULL,
	mxn_spec_server_resynchronize
};

MX_RECORD_FIELD_DEFAULTS mxn_spec_server_record_field_defaults[] = {
	MX_RECORD_STANDARD_FIELDS,
	MXN_SPEC_SERVER_STANDARD_FIELDS
};

long mxn_spec_server_num_record_fields
		= sizeof( mxn_spec_server_record_field_defaults )
			/ sizeof( mxn_spec_server_record_field_defaults[0] );

MX_RECORD_FIELD_DEFAULTS *mxn_spec_server_rfield_def_ptr
			= &mxn_spec_server_record_field_defaults[0];

MX_EXPORT mx_status_type
mxn_spec_server_create_record_structures( MX_RECORD *record )
{
	const char fname[] = "mxn_spec_server_create_record_structures()";

	MX_SPEC_SERVER *spec_server;

	/* Allocate memory for the necessary structures. */

	spec_server = (MX_SPEC_SERVER *) malloc( sizeof(MX_SPEC_SERVER) );

	if ( spec_server == (MX_SPEC_SERVER *) NULL ) {
		return mx_error( MXE_OUT_OF_MEMORY, fname,
		"Can't allocate memory for MX_SPEC_SERVER structure." );
	}

	/* Now set up the necessary pointers. */

	record->record_class_struct = NULL;
	record->record_type_struct = spec_server;
	record->class_specific_function_list = NULL;

	spec_server->record = record;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxn_spec_server_finish_record_initialization( MX_RECORD *record )
{
	MX_SPEC_SERVER *spec_server;

	spec_server = (MX_SPEC_SERVER *) record->record_type_struct;

	spec_server->socket = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxn_spec_server_open( MX_RECORD *record )
{
	const char fname[] = "mxn_spec_server_open()";

	MX_LIST_HEAD *list_head;
	MX_SPEC_SERVER *spec_server;
	MX_SOCKET *server_socket;
	long dimension_array[1];
	char buffer[20];
	mx_status_type mx_status;

	list_head = mx_get_record_list_head_struct( record );

	if ( list_head == (MX_LIST_HEAD *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"The MX_LIST_HEAD pointer for record '%s' is NULL.",
			record->name );
	}

	spec_server = (MX_SPEC_SERVER *) record->record_type_struct;

	if ( spec_server == (MX_SPEC_SERVER *) NULL ) {
		return mx_error( MXE_CORRUPT_DATA_STRUCTURE, fname,
		"MX_SPEC_SERVER pointer for record '%s' is NULL.",
			record->name );
	}

	mx_status = mx_tcp_socket_open_as_client( &server_socket,
			spec_server->hostname, spec_server->port, 0,
			MX_SOCKET_DEFAULT_BUFFER_SIZE );

	switch( mx_status.code ) {
	case MXE_SUCCESS:
		spec_server->socket = server_socket;
		break;

	case MXE_NETWORK_IO_ERROR:
		spec_server->socket = NULL;

		return mx_error( MXE_NETWORK_IO_ERROR, fname,
"The Spec server at port %d on the computer '%s' is either not running "
"or is not working correctly.  "
"You can try to fix this by restarting the Spec server.",
			spec_server->port, spec_server->hostname );

		break;

	default:
		spec_server->socket = NULL;

		return mx_error( mx_status.code, fname,
"An unexpected error occurred while trying to connect to the Spec server "
"at port %d on the computer '%s'.", spec_server->port, spec_server->hostname);

		break;
	}

	/* Check to see if the Spec server is working by asking for
	 * something innocuous, namely, whether or not the Spec server
	 * is in simulate mode.
	 */

	dimension_array[0] = sizeof(buffer);

	mx_status = mx_spec_get_array( record, "status/simulate",
					SV_STRING, 1, dimension_array,
					buffer );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	MX_DEBUG(-2,("%s: status/simulate buffer = '%s'",
		fname, buffer ));

	return mx_status;
}

MX_EXPORT mx_status_type
mxn_spec_server_close( MX_RECORD *record )
{
	MX_SPEC_SERVER *spec_server;
	MX_SOCKET *server_socket;
	mx_status_type mx_status;

	spec_server = (MX_SPEC_SERVER *) record->record_type_struct;

	server_socket = spec_server->socket;

	if ( server_socket != NULL ) {
		mx_status = mx_socket_close( server_socket );

		if ( mx_status.code != MXE_SUCCESS )
			return mx_status;
	}
	spec_server->socket = NULL;

	return MX_SUCCESSFUL_RESULT;
}

MX_EXPORT mx_status_type
mxn_spec_server_resynchronize( MX_RECORD *record )
{
	mx_status_type mx_status;

	mx_status = mxn_spec_server_close( record );

	if ( mx_status.code != MXE_SUCCESS )
		return mx_status;

	mx_status = mxn_spec_server_open( record );

	return mx_status;
}

